@@ -1,5 +1,5 @@
 /*
- * Copyright 2000-2003 by Hans Reiser, licensing governed by 
+ * Copyright 2000-2004 by Hans Reiser, licensing governed by 
  * reiserfsprogs/README
  */
   
@@ -616,7 +616,7 @@ int main (int argc, char ** argv)
 	/* only one non-option argument is permitted */
 	print_usage_and_exit();
 
-    if (is_mounted (argv[optind]))
+    if (misc_device_mounted(argv[optind], 0))
 	reiserfs_panic ("%s seems mounted, umount it first\n", argv[optind]);
   
     fd = open (argv[optind], O_RDWR | O_LARGEFILE);
